from datetime import datetime


def current_time_as_str():
    """
    read current time and process it to string to be used in a file name.
    :return: str, current time value
    """
    dateNtime = str(datetime.now()).replace(" ", '')
    date = dateNtime[2:10].replace("-", "")
    time = dateNtime[10:-10].replace(":", "")

    return date + "_" + time


def flatten_list(txts):
    """
    flatten list of texts by concatenating
    :param txts: str, list of texts
    :return: str, flattened text as string
    """
    flatten = str()
    for txt in txts:
        flatten = flatten + txt
    return flatten



import logging

import win32com.client
import os
import codecs
import re
import pandas as pd
from util import current_time_as_str


def typodelete(txt_id, txt_list, output_dir, save=True):
    """
    Detect typo in a list of texts using MSword and delete the typos
    :param output_dir: str, output directory to save the processed text
    :param txt_id: unique id of texts
    :param txt_list: list of texts
    :param save: boolean, if True, save the result file to excel
    :return: df, which contains raw text, typos, and typo deleted text
    """

    logging.info("Processing the raw input: typo deletion . . .")
    # make df to store results
    df_column = ['raw', 'typo', 'processed']
    output_df = pd.DataFrame(index=txt_id, columns=df_column)

    # save the raw text into ms word form to use ms word typo corrector
    current_time = current_time_as_str()
    file_path = os.path.abspath(current_time + ".txt")
    # logging.info("Concatenating files into a single .txt . . .")
    with codecs.open(file_path, 'w', encoding='utf-8') as f:
        index = 0
        for text in txt_list:
            index += 1
            f.write(text)
            if len(txt_list) > index:  # do not add \n at the end of the file
                f.write('\n')

    # open the ms word document
    msword = win32com.client.DispatchEx('Word.Application')
    # logging.info("Loading .txt into MS Word . . .")
    doc = msword.Documents.Open(file_path)
    assert int(doc.Paragraphs.Count) == len(txt_list)

    # for each text(Paragraph in word), detect error
    # logging.info("Processing text started . . .")
    for i, p in enumerate(doc.Paragraphs):  # process per text
        tab_removed_raw = re.sub("\t", " ", txt_list[i])  # remove all tabs to store the data as tsv file
        typos = p.Range.SpellingErrors
        item = []
        typo_list = []

        for typo in typos:  # make typo list
            typo_list.append(str(typo))

        # delete typos in text
        pattern = re.compile(r'\b(?:%s)\b' % '|'.join(typo_list))
        processed = re.sub(pattern, '', tab_removed_raw)

        item.append(tab_removed_raw)
        item.append(typo_list)
        item.append(processed)
        output_df.loc[txt_id[i]] = item

        assert "\t" not in tab_removed_raw, f"File {txt_id[i]} contains tab, Please remove all tabs in the file for generating tsv file. "

    # close and delete temporary files
    doc.Close(0)
    msword.Quit(0)
    os.remove(file_path)

    if save:
        file_path = output_dir + '/' + "processed_data" + ".tsv"
        logging.info("Saving processed text file as %s . . .", file_path)
        output_df.to_csv(file_path, encoding='utf-8', sep='\t')

    logging.info("Typo deletion completed")

    return output_df




from data_processor import typodelete
from data_reader import read_texts_into_lists
import argparse
import logging
from ld_analyser import tokenize_n_make_ld_matrix
import warnings
import numpy as np
import pandas as pd
import sys

from util import current_time_as_str

warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
logging.basicConfig(level=logging.INFO)
logging.getLogger("tensorflow").setLevel(logging.CRITICAL)
logging.getLogger("stanza").setLevel(logging.WARNING)

if __name__ == '__main__':
    # parse args
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--inputdir", required=True,
                        help="Path to the directory which includes plain text files to be analysed")
    parser.add_argument("-t", "--tokenizer", nargs='+', default=["okt"],
                        help="Tokenizers: (okt, komoran, mecab, kkma, hannanum, stanza). You can give multiple tokenizers ex) -t kkma komoran")
    parser.add_argument("-a", "--all", action="store_true",
                        help="""Do analysis on all possible combinations on the given tokenizer sets (function word True & False x parallel analysis True & False)
                             Note that the argument functionwords=false is not provided in stanza.
                             Argument --functionwords and --parallel are ignored if -a set true """)
    parser.add_argument("-f", "--functionwords", action='store_true',
                        help="""Do analysis on both function and content words. If this argument not given, do analysis only on content words.
                                Note that functionwords=false in not provided in stanza. """)
    parser.add_argument("-p", "--parallel", action='store_true', help="do parallel analysis.")
    parser.add_argument("-o", "--outputdir", default='result', help="path to store output files (log, tsv files with ld values")
    parser.add_argument("-no-typo-removal", "--notyporemoval", action='store_true', help="Note: Windows OS with Microsoft Office is required for this function.")
    args = parser.parse_args()


    # set logger
    a_logger = logging.getLogger()
    a_logger.setLevel(logging.DEBUG)
    log_file = args.outputdir + '/' + "log_" + current_time_as_str() + ".log"
    output_file_handler = logging.FileHandler(log_file)
    a_logger.addHandler(output_file_handler)

    if "stanza" in args.tokenizer and not args.functionwords: # if the user wants to exclude functionwords in stanza
        if not args.all:
            raise ValueError("Stanza does not provide functionwords=False. Please give functionwords argument for stanza tokenizer by adding -f to the command")

    # configuration information (show as log)
    logging.info("-----------Configuration----------")
    logging.info("Selected Tokenizer = %s", args.tokenizer)
    if args.all:
        logging.info("Four different analysis will be done for the selected tokenizer(s)")
        logging.info("(functionword True, False) x (Parallel analysis True, False)")
    else:
        logging.info("Include Function Words = %s", args.functionwords)
        logging.info("Parallel Analysis = %s", args.parallel)
    if args.notyporemoval:
        logging.info("Typo removal function is off: No typos will be removed.")
    logging.info("----------------------------------")

    # read and process text

    if args.notyporemoval:
        txt_id, text_list = read_texts_into_lists(args.inputdir, remove_num=False)
        data_df = pd.DataFrame(index=txt_id, columns=['processed'])
        data_df['processed'] = text_list
    else:
        txt_id, text_list = read_texts_into_lists(args.inputdir)
        data_df = typodelete(txt_id, text_list, args.outputdir)

    # tokenize and analyse
    if args.all:
        f_options = [True, False]
        p_options = [True, False]
        for tokenizer in args.tokenizer:
            for f in f_options:
                for p in p_options:
                    config = "Tokenizer: {}, Include Function Words: {}, Parallel Analysis: {}".format(tokenizer, f, p)
                    logging.info("\n\n\n================ %s =================", config)
                    tokenize_n_make_ld_matrix(data=data_df, tokenizer=tokenizer,
                                              include_function_words=f,
                                              parallel_analysis=p, output_dir=args.outputdir)


    else:
        for tokenizer in args.tokenizer:
            logging.info("\n\n\n================ tokenizer %s =================", tokenizer)
            tokenize_n_make_ld_matrix(data=data_df, tokenizer=tokenizer,
                                      include_function_words=args.functionwords, parallel_analysis=args.parallel, output_dir=args.outputdir)

    logging.info("FINISHED")



from konlpy.tag import Okt, Komoran, Kkma, Hannanum
import mecab_ko as MeCab
try:
    import stanza
except ImportError:
    pass

OKT_STOPWORDS = ["Punctuation", "Foreign", "Alpha", "Number", "Unknown", "KoreanParticle", "Hashtag", "ScreenName",
                 "Email", "URL"]
KOMORAN_STOPWORDS = ["SF", "SE", "SS", "SP", "SO", "SW", "SH", "SL", "SN", "NF", "NV", "NA"]

MECAB_STOPWORDS = ["SF", "SE", "SSO", "SSC", "SC", "SY", "SH", "SL", "SN"]

KKMA_STOPWORDS = ["SF", "SE", "SS", "SP", "SO", "SW", "OH", "OL", "ON", "UN"]

HANNANUM_STOPWORDS = ['S', 'F']

STANZA_STOPWORDS = ['PUNCT', 'SYM', 'X']

OKT_FUNCTIONWORDS = ["Josa", "PreEomi", "Eomi", "Suffix"]

KOMORAN_FUNCTIONWORDS = ["JKS", "JKC", "JKG", "JKO", "JKB", "JKV", "JKQ", "JC", "JX", "EP", "EF", "EC", "ETN", "ETM",
                         "XPN", "XSN", "XSV", "XSA"]

MECAB_FUNCTIONWORDS = ["JKS", "JKC", "JKG", "JKO", "JKB", "JKV", "JKQ", "JC", "JX", "EP", "EF", "EC", "ETN", "ETM",
                       "XPN", "XSN", "XSV", "XSA"]

KKMA_FUNCTIONWORDS = ["JKS", "JKC", "JKG", "JKO", "JKM", "JKI", "JKQ", "JC", "JX", "EPH", "EPT", "EPP", "EFN", "EFQ",
                      "EFO", "EFA", "EFI", "EFR", "ECE", "ECS", "ECD", "ETN", "ETD", "XPN", "XPV", "XSN", "XSV", "XSA"]

HANNANUM_FUNCTIONWORDS = ["J", "E", "X"]


def remove_pos(token_pos_tuple, pos_list):
    """
    Remove given POSs in the tokenized tuple
    :param token_pos_tuple: list of tuple consisting of token and POS [('열심히', 'Adverb'), ('코딩', 'Noun')...]
    :param pos_list: list of POS to be removed
    :return: cleaned tuple of ('token', 'Part-Of-Speech') as list
    """
    pos_tuple_cleaned = []

    for index, pair in enumerate(token_pos_tuple):
        if pair[1] not in pos_list:
            pos_tuple_cleaned.append(pair)

    return pos_tuple_cleaned


def tokenize(tokenizer, text):
    """
    tokenize sequences using konlpy tokenizer.
    :param tokenizer: str, possible options: (okt, komoran, mecab, kkma, hannanum, stanza)
    :param text: str, raw text
    :return: tuple (pos_tuple_all, pos_tuple_cleaned, tokens_cleaned)
            where pos_tuple_all consists of tuple ('token', 'Part-Of-Speech') of all raw tokens (including stopwords like punctuation, numbers, URL ...)
                  pos_tuple_cleaned consists of tuple ('token', 'Part-Of-Speech') of contents words (+ function words if the param include_function_words=True)
                  tokens_cleaned is a list of stopword removed tokens (if the param include_function_words=True, function words are also included)
    """

    if tokenizer == 'okt':
        tagger = Okt()
        stopwords = OKT_STOPWORDS
    elif tokenizer == 'komoran':
        tagger = Komoran()
        stopwords = KOMORAN_STOPWORDS
    elif tokenizer == 'mecab':
        tagger = MeCab.Tagger()
        stopwords = MECAB_STOPWORDS
    elif tokenizer == 'kkma':
        tagger = Kkma()
        stopwords = KKMA_STOPWORDS
    elif tokenizer == 'hannanum':
        tagger = Hannanum()
        stopwords = HANNANUM_STOPWORDS
    elif tokenizer == 'stanza':
        tagger = stanza.Pipeline('ko', processors='tokenize,pos', package='gsd')
        stopwords = STANZA_STOPWORDS
    else:
        raise ValueError("tokenizer must be one of these options: (okt, komoran, mecab, kkma, hannanum, stanza)")

    # tokenize
    if tokenizer == 'stanza':
        doc = tagger(text)
        pos_tuple_all = [(word.text, word.upos) for sent in doc.sentences for word in sent.words]
    elif tokenizer == 'mecab':
        tmp_mecab = tagger.parse(text)
        tmp_mecab = tmp_mecab.split("\n")
        tmp_mecab = tmp_mecab[:-2] # mecab_ko has an [EOS] symbol. remove this
        tmp_mecab = [item.split("\t") for item in tmp_mecab]
        pos_tuple_all = [(item[0], item[1].split(",")[0]) for item in tmp_mecab]
    else:
        pos_tuple_all = tagger.pos(text)

    # remove stopwords
    pos_tuple_cleaned = remove_pos(pos_tuple_all, pos_list=stopwords)

    # separate lists for tokens
    tokens_cleaned = [item[0] for item in pos_tuple_cleaned]

    return pos_tuple_all, pos_tuple_cleaned, tokens_cleaned


def remove_function_words(pos_tuple, tokenizer):
    """
    Remove function words from tokenized pos tuple
    include this if main argument functionwords=False
    :param pos_tuple: ('token', 'Part-Of-Speech')
    :param tokenizer: str, available options: okt, komoran, mecab, kkma, hannanum
    :return:
    """

    functionwords = []

    if tokenizer == 'okt':
        functionwords = OKT_FUNCTIONWORDS
    elif tokenizer == 'komoran':
        functionwords = KOMORAN_FUNCTIONWORDS
    elif tokenizer == 'mecab':
        functionwords = MECAB_FUNCTIONWORDS
    elif tokenizer == 'kkma':
        functionwords = KKMA_FUNCTIONWORDS
    elif tokenizer == 'hannanum':
        functionwords = HANNANUM_FUNCTIONWORDS

    pos_tuple_cleaned = remove_pos(pos_tuple, pos_list=functionwords)

    # separate lists for tokens
    tokens_cleaned = [item[0] for item in pos_tuple_cleaned]

    return pos_tuple_cleaned, tokens_cleaned



import os
import logging
from string import digits


def read_text_file(file_path):
    with open(file_path, 'r', encoding='utf-8') as f:
        txt = f.read()
        txt = txt.replace("\n", " ")
    return txt


def read_texts_into_lists(path, remove_num=True):
    """
    read all plain text files (.txt) in the path and return text id (file name) and text contents as list
    :param remove_num: bool, if set True, remove numbers in the text.
                        default is set True. (Processing numbers are not consistent for korean tokenizers, so remove nums beforehand.)
    :param path: str, directory to the input files
    :return: tuple (txt_id, text_list)
                where (txt_id) is a list of file names
                      (text_list) is a list of file contents as string
    """
    owd = os.getcwd()
    os.chdir(path)
    text_list = list()
    txt_id = list()
    for file in os.listdir():
        if file.endswith(".txt"):
            txt = read_text_file(file)
            if len(txt) < 1:
                logging.info("%s is empty file. Skipped", file)
                continue
            txt_id.append(file)
            txt = read_text_file(file)
            text_list.append(txt)

    os.chdir(owd)

    if remove_num:   # remove number
        text_list = [txt.translate(str.maketrans('', '', digits)) for txt in text_list]

    logging.info("%s files read successfully.", len(txt_id))

    return txt_id, text_list



from korean_tokenizer import tokenize, remove_function_words
from taaled import parallel, lexdiv
import logging
from util import current_time_as_str


def tokenize_n_make_ld_matrix(data, tokenizer, include_function_words, parallel_analysis, output_dir, mx=200):
    """
    Tokenize and calculate all files in the df data and write output as tsv
    (This code includes partial modification of TAALED package source code)
    :param output_dir: str, output directory to store result files
    :param data: df, dataframe with three columns: text id, raw text, processed (typo removed) text, where df index is text file name
    :param tokenizer: str, possible options: (okt, komoran, mecab, kkma, hannanum, stanza)
    :param include_function_words: bool
                                    if set True: tokenize content + function words
                                    if set False: tokenize only content words
    :param parallel_analysis: bool
                        if set True: do parallel analysis
                        if set False: no parallel analysis
    :param mx: int, minimum length of a text for parallel analysis
    :return: none
    """

    # filter out stanza with include_function_words=False
    if tokenizer == 'stanza' and not include_function_words:
        return

    logging.info("Start LD analysis . . .")
    # indexes
    loi = ["ntokens", "ntypes", "mtld", "mtldo", "mattr", "ttr", "rttr", "lttr", "maas", "msttr", "hdd"]

    # set file name
    if include_function_words:
        file_name = output_dir + "/" + tokenizer + "_all_words.tsv"
        if parallel_analysis:
            file_name = output_dir + "/" + tokenizer + "_all_words_prll.tsv"
    else:
        file_name = output_dir + "/" + tokenizer + "_content_only.tsv"
        if parallel_analysis:
            file_name = output_dir + "/" + tokenizer + "_content_only_prll.tsv"

    outf = open(file_name, "w", encoding='utf-8')
    if parallel_analysis:
        outf.write("filename" + '\t' + "length" + '\t' + '\t'.join(loi))
    else:
        outf.write("filename" + '\t' + '\t'.join(loi))

    text_id = data.index
    text_processed = data['processed']
    skipped = 0
    skippedl = []

    for i, text in enumerate(text_processed):
        _, pos_tuple, tokens_cleaned = tokenize(tokenizer, text)
        if not include_function_words:
            _, tokens_cleaned = remove_function_words(pos_tuple, tokenizer)
        if len(tokens_cleaned) < 1:  # nothing to analysis
            logging.info("%s has no analysable tokens. Skipping", text_id[i])
            skipped += 1
            skippedl.append(text_id[i])
            continue
        if parallel_analysis:
            if len(tokens_cleaned) < mx:  # in case the text is too short for parallel analysis
                skipped += 1
                skippedl.append(text_id[i])
                continue
            ld_lists = parallel(text=tokens_cleaned, clss=True, functd=None, funct=lexdiv, loi=loi, mx=mx).ldvals
            for length in ld_lists:  # iterate through text slices
                outl = [text_id[i], str(length)]  # list of items to write, will add each index below
                for index in loi:  # iterate through index list:
                    outl.append(str(ld_lists[length][index]))  # add index to outr list (in same order as loi list)
                outf.write("\n" + '\t'.join(outl))  # write row to file
        else:  # no parallel analysis
            ldout = lexdiv(tokens_cleaned).vald  # get dictionary version of lexical diversity output
            outl = [text_id[i]]  # list of items to write, will add each index below
            for index in loi:  # iterate through index list:
                outl.append(str(ldout[index]))  # add index to outr list (in same order as loi list)
            outf.write("\n" + '\t'.join(outl))  # write row to file

    outf.flush()
    outf.close()

    logging.info("Analysis on %s files completed successfully", len(text_id) - skipped)
    logging.info("The result is saved as: %s", file_name)

    if skipped:
        logging.info("%s files are skipped due to length problem", skipped)
        logging.info("List of skipped files: ")
        logging.info("%s", skippedl)
